/***************************************************************************
 * jack-fifo
 * 
 * This program's purpose is to allow a program using /dev/dsp as input
 * to be able to use Jack Audio Connection Kit, which can then distribute
 * the sound to other sinks as well.
 * 
 * It assumes /dev/dsp was set up to use 16 bit little endian stereo.  If you 
 * application uses something different, you will need to make the appropriate 
 * modifications.
 *
 * Written:
 *  June 2010
 *  by Rob Frohne
 *  e-mail: Rob "dot" Frohne "at sign" wallawalla "dot" edu
 * 
 ****************************************************************************/

/* To build this program, you must have jack audio connection kit installed.
 * At the command prompt:
 * $ gcc jack-fifo.c -ljack -o jack-fifo
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <jack/jack.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>

 
#define PIPE_NAME		"/tmp/jackfifo" //Default fifo

/* global jack variables. */
jack_client_t *test_client;
jack_port_t *audio_input_port_left, *audio_input_port_right;
int pipe_handle = 0;

/* To use this command, you need to $ jack-fifo -v -f fifo-name
 * assuming those are the options you want.
 * Options are:
 * -? for help
 * -v or --verbose for verbose response.
 * -f or --name of the fifo to use or create and use
 */
/* Flag set by ‘--verbose’. */
static int verbose_flag;
// Specify options and their syntax
static struct option long_options[] =
{
	/* These options set a flag. */
	{"verbose", no_argument,       &verbose_flag, 1},
	{"brief",   no_argument,       &verbose_flag, 0},
	/* These options don't set a flag.
	We distinguish them by their indices. */
	{"help",     no_argument,      0, '?'},
	{"fifo",   required_argument,       0, 'f'},
	{0, 0, 0, 0}
};

/* Close things opened. */
void cleanup(void) {
	if(pipe_handle != 0) close(pipe_handle);
	if(test_client != NULL) {
		jack_deactivate(test_client);	
		jack_client_close(test_client);
	}
}

/* This is the function that is called when and if jackd shuts down. */
void
jack_shutdown (void *arg)
{
	fprintf (stderr, "JACK shutdown\n");
	cleanup();
	abort();
}
 
/* This is the callback process that gets data from jack.*/	
int process(jack_nframes_t number_of_frames, void *arg)
{
	signed short x_left, x_right, sound_buffer[number_of_frames*4];
	int * pipe_handle_ptr = (int *) arg;
	static int error_count;
	
	jack_default_audio_sample_t *sample_buffer_left = 
	(jack_default_audio_sample_t *) jack_port_get_buffer(audio_input_port_left, number_of_frames);
 
	jack_default_audio_sample_t *sample_buffer_right = 
	(jack_default_audio_sample_t *) jack_port_get_buffer(audio_input_port_right, number_of_frames);
      
	jack_nframes_t i;
	//This is where you get your data in the format that the program using /dev/dsp needs it.
	for(i = 0; i < number_of_frames; i++){
		x_left = (signed short) rint(sample_buffer_left[i]*32767);  //Convert from double to signed short.
		x_right = (signed short) rint(sample_buffer_right[i]*32767);
		// We need to write the data (left channel, right channel, left channel, right channel,...)
		sound_buffer[2*i] = x_left; 
		sound_buffer[2*i+1] = x_right;
	}
	/* This is where you write the data to the pipe. Note that if the program you are accomadating
	 * has used ioctl to set up some other format than 16 bit stereo interleved for the data, you 
	 * probably need to modify both the section above and the size of the data in the write statement
	 * below.  For more information on /dev/dsp, see this site:
	 * http://www.oreilly.de/catalog/multilinux/excerpt/ch14-05.htm
	 */
	if(write(*pipe_handle_ptr, &sound_buffer, 4*number_of_frames) < 0) 
		if( error_count++ < 100) fprintf(stderr, "Could not write to the fifo. (error %s)\n", strerror( errno ));
	return 0;
}


int main (int argc, char *argv[]){
	
	int error;
	
	char * fifo_name = PIPE_NAME;  
	
	
	// Iterate over options and handle each one as appropriate
	int c;
	// The function getopt_long stores the option index here.
	int option_index = 0;

	while ( (c = getopt_long ( argc, argv, "v?f:",
	                          long_options, &option_index )) != -1 )
	{
		switch ( c )
		{
			case 0:
				// If this option set a flag, do nothing else now.
				if ( long_options[option_index].flag != 0 )
					break;

			case 'v':
				verbose_flag = 1;
				break;

			case 'f':
				fifo_name = optarg;
				break;
				
			case '?':

			default :
				printf(
				"Options are: \n"
				"-v or --verbose for verbose output\n"
				"-f or --fifo-name (to use and possibly create)\n\n"
				"Typing EOF (ctrl-D) (ctrl-capitol D, to be exact) will end the program.\n");
				exit ( 0 );
		}
	}
	
	/* Setup the Fifo */
	//Make the fifo.
	if( mkfifo(fifo_name, S_IRUSR|S_IWUSR|S_IROTH|S_IWOTH) != 0) {
		if(errno != EEXIST){ 
			fprintf(stderr, "Could not make the fifo.  Error: %i\n",errno);
			cleanup();
			return 1;
		}
		else if((verbose_flag)) fprintf(stderr,"The fifo, %s already exists; we will use it.\n",fifo_name);
	}
	else if(verbose_flag) fprintf(stderr, "Made the fifo, %s.\n",fifo_name);

  	// Open the pipe for writing.
	pipe_handle = open(fifo_name, O_NONBLOCK|O_RDWR);//O_NONBLOCK|O_WRONLY
	if (pipe_handle < 0){
			fprintf(stderr, "Could not open the fifo %s.  Error: %s\n", fifo_name, strerror( errno ));
			if(errno == ENXIO) fprintf(stderr, "You must start the program that reads the fifo before\n"
			"you can start jack-fifo.  Go do that and try again.\n");
			/* See:  http://www.ia.pw.edu.pl/~wujek/dokumentacja/gnu/libc/libc_8.htm
			 * You get an ENXIO error if O_NONBLOCK and O_WRONLY are both set in the flags argument,
			 * the file named by filename is a FIFO (see section Pipes and FIFOs), and no process has 
			 * the file open for reading.  To get around this we use the O_RDWR flag, so the above
			 * error shouldn't happen.
			 */
			cleanup();
			return 1;
		}
		if(verbose_flag) fprintf(stderr,"Opened the fifo.  Handle is: %i\n", pipe_handle);
	//Create a new jack client, then make sure everything went ok.
	test_client = jack_client_open("Test Client",(jack_options_t)(!JackServerName),NULL);
	if (test_client == 0) {
    	fprintf(stderr,"Cannot connect to the jackd as a client.\n");
    	cleanup();
    	return 1;
  	}
  	  	/* Set up Jack */
  	//Set up the jack shutdown routine in case we want to do something special on shutdown of jack.
  	jack_on_shutdown (test_client, jack_shutdown, 0);

	//Create and register new audio input ports.
	audio_input_port_left = jack_port_register(test_client, "Audio FIFO Port_left", 
						JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	if (audio_input_port_left == NULL) {
		fprintf(stderr, "Error: jack_port_register returned NULL for left channel.\n");
		cleanup();
		return 1;
	}
		
	audio_input_port_right = jack_port_register(test_client, "Audio FIFO Port_right", 
						JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	if (audio_input_port_right == NULL) {
		fprintf(stderr, "Error: jack_port_register returned NULL for right channel.\n");
		cleanup();
		return 1;
	}
	
	//Tell the jackd server what function call when it wants more audio data.
	if((error = jack_set_process_callback(test_client, process, &pipe_handle)) != 0) { 
		fprintf(stderr, "Jack could not set the callback, (error %i).\n", error);
		cleanup();
		return 1;
	}
	
	//Tell jack it's ok to start asking us for audio data.
	if((error = jack_activate(test_client)) != 0) {
		fprintf(stderr, "Jack could not activate the client (error %i).\n", error);
		cleanup();
		return 1;
	}
	else if(verbose_flag) fprintf(stderr,"Activated client.\n");
	
	//Connect the ports.	
	if (jack_connect (test_client,"system:capture_1", jack_port_name (audio_input_port_left))) {
        fprintf (stderr, "cannot connect port: system:capture_1\n");
    }
    else if(verbose_flag) fprintf(stderr, "Connected to port: system:capture_1\n");
	if (jack_connect (test_client,"system:capture_2", jack_port_name (audio_input_port_right))) {
        fprintf (stderr, "cannot connect port: system:capture_2\n");
    }
    else if(verbose_flag) fprintf(stderr, "Connected to port: system:capture_2\n");
	
	while(getchar() != EOF){
		//Quit on ctrl-D.
	}
	
	/* Close things opened. */
	cleanup();	

	return 0;
}
 
