/***************************************************************************
 * jack-fifo
 * 
 * This program's purpose is to allow a program using /dev/dsp as input
 * to be able to use Jack Audio Connection Kit, which can then distribute
 * the sound to other sinks as well.
 * 
 * It assumes /dev/dsp was set up to use 16 bit little endian stereo.  If you 
 * application uses something different, you will need to make the appropriate 
 * modifications.
 *
 * Written:
 *  June 2010
 *  by Rob Frohne
 *  e-mail: Rob "dot" Frohne "at sign" wallawalla "dot" edu
 * 
 ****************************************************************************/

/* To build this program, you must have jack audio connection kit installed.
 * At the command prompt:
 * $ gcc -jack-fifo.c -ljack -o jack-fifo
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <jack/jack.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <getopt.h>
 
#define PIPE_NAME1		"/tmp/jack1" //Default fifo
#define PIPE_NAME2		"/tmp/jack2" //Default fifo
#define PIPE_NAME3		"/tmp/jack3" //Default fifo
#define NUMBER_PIPES 3

typedef struct {
	int pipe_handle[NUMBER_PIPES];
} process_args;


jack_port_t *audio_input_port_left[3], *audio_input_port_right[NUMBER_PIPES];

/* To use this command, you need to $ jack-fifo -v -f fifo-name
 * assuming those are the options you want.
 * Options are:
 * -? for help
 * -v or --verbose for verbose response.
 * -f or --name of the fifo to use or create and use
 */
/* Flag set by ‘--verbose’. */
static int verbose_flag;
// Specify options and their syntax
static struct option long_options[] =
{
	/* These options set a flag. */
	{"verbose", no_argument,       &verbose_flag, 1},
	{"brief",   no_argument,       &verbose_flag, 0},
	/* These options don't set a flag.
	We distinguish them by their indices. */
	{"help",     no_argument,      0, '?'},
	{"fifo",   required_argument,       0, 'f'},
	{0, 0, 0, 0}
};

void cleanup( int * pipe_handle, jack_client_t *test_client ) {
	int j;
		/* Close things opened. */
	for(j=0; j<NUMBER_PIPES; j++) {
		if(pipe_handle[j] != 0) close(pipe_handle[j]);
	}
	if(test_client != NULL) {
		jack_deactivate(test_client);	
		jack_client_close(test_client);
	}
} 	
 
 
int process(jack_nframes_t number_of_frames, void *arg)
{
	signed short x_left[NUMBER_PIPES], x_right[NUMBER_PIPES], sound_buffer[number_of_frames*4]; //Perhaps should be number_of_frames*2.
	process_args * arg1 = ( process_args *) arg;
	int j;
	static int error_count = 0;
	
	jack_default_audio_sample_t *sample_buffer[6];
      
	jack_nframes_t i;
	
	//This is where you get your data in the format that the program using /dev/dsp needs it.
	for(j=0; j < NUMBER_PIPES; j++){
		sample_buffer[2*j] = (jack_default_audio_sample_t *) jack_port_get_buffer(audio_input_port_left[j], number_of_frames);
		sample_buffer[2*j+1] = (jack_default_audio_sample_t *) jack_port_get_buffer(audio_input_port_right[j], number_of_frames);
		for(i = 0; i < number_of_frames; i++){
			x_left[j] = (signed short) rint(sample_buffer[2*j][i]*32767);  //Convert from double to signed short.
			x_right[j] = (signed short) rint(sample_buffer[2*j+1][i]*32767);
			// We need to write the data (left channel, right channel, left channel, right channel,...)
			sound_buffer[2*i] = x_left[j]; 
			sound_buffer[2*i+1] = x_right[j];
		}
		/* This is where you write the data to the pipe. Note that if the program you are accomadating
		 * has used ioctl to set up some other format than 16 bit stereo interleved for the data, you 
		 * probably need to modify both the section above and the size of the data in the write statement
		 * below.  For more information on /dev/dsp, see this site:
		 * http://www.oreilly.de/catalog/multilinux/excerpt/ch14-05.htm
		 */
		if(write(arg1->pipe_handle[j], sound_buffer, 4*number_of_frames) < 0) 
			if( error_count++ < 15) fprintf(stderr, "Could not write to the fifo. (error %i)\n", errno);
	}
	return 0;
}


int main (int argc, char *argv[]){
	jack_client_t *test_client;
	int j, error, pipe_handle[NUMBER_PIPES];
	char *fifo_name[NUMBER_PIPES] = {PIPE_NAME1, PIPE_NAME2, PIPE_NAME3};
	char v_jack_port_name[256];  
	
	
	// Iterate over options and handle each one as appropriate
	int c;
	// The function getopt_long stores the option index here.
	int option_index = 0;

	while ( (c = getopt_long ( argc, argv, "v?",
	                          long_options, &option_index )) != -1 )
	{
		switch ( c )
		{
			case 0:
				// If this option set a flag, do nothing else now.
				if ( long_options[option_index].flag != 0 )
					break;

			case 'v':
				verbose_flag = 1;
				break;

			//case 'f':
				//fifo_name = optarg;
				//break;
				
			case '?':

			default :
				printf(
				"Options are: \n"
				"-v or --verbose for verbose output\n"
				"-f or --fifo-name (to use and possibly create)\n\n"
				"Typing EOF (ctrl-D) (ctrl-capitol D, to be exact) will end the program.\n");
				exit ( 0 );
		}
	}
	
	//Create a new jack client, then make sure everything went ok.
	test_client = jack_client_open("Test Client",(jack_options_t)(!JackServerName),NULL);
	if (test_client == 0) {
    	fprintf(stderr,"Cannot connect to the jackd as a client.\n");
    	cleanup( pipe_handle, test_client );
    	return 1;
  	}
  	

	for( j=0; j<NUMBER_PIPES; j++) {
		//Make the fifo.
		if( mkfifo( fifo_name[j], S_IRUSR|S_IWUSR|S_IROTH|S_IWOTH) != 0) {
			if(errno != EEXIST){ 
				fprintf(stderr, "Could not make the fifo %s.  Error: %i\n", fifo_name[j], errno);
				cleanup( pipe_handle, test_client );
				return 1;
			}
			else if((verbose_flag)) fprintf(stderr,"The fifo, %s already exists; we will use it.\n",fifo_name[j]);
		}
		else if(verbose_flag) fprintf(stderr, "Made the fifo, %s.\n",fifo_name[j]);
	
	  	// Open the pipe for writing.
		pipe_handle[j] = open( fifo_name[j], O_RDWR|O_NONBLOCK);//O_NONBLOCK|O_WRONLY
		if (pipe_handle[j] < 0){
				fprintf(stderr, "Could not open the fifo %s.  Error: %i\n", fifo_name[j], errno);
				if(errno == ENXIO) fprintf(stderr, "You must start the program that reads the fifo before\n"
				"you can start jack-fifo.  Go do that and try again.\n");
				/* See:  http://www.ia.pw.edu.pl/~wujek/dokumentacja/gnu/libc/libc_8.htm
				 * You get an ENXIO error if O_NONBLOCK and O_WRONLY are both set in the flags argument,
				 * the file named by filename is a FIFO (see section Pipes and FIFOs), and no process has 
				 * the file open for reading.
				 */
				cleanup( pipe_handle, test_client );
				return 1;
			}
			if(verbose_flag) fprintf(stderr,"Opened the fifo %s.  Handle is: %i\n", fifo_name[j], pipe_handle[j]);
		
	  	/* Set up Jack */
		//Create and register new audio input ports.
		
		sprintf(v_jack_port_name , "Audio FIFO Port_left_ %i", j);
		audio_input_port_left[j] = jack_port_register(test_client, v_jack_port_name, 
							JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
		if (audio_input_port_left[j] == NULL) {
			fprintf(stderr, "Error: jack_port_register returned NULL for left channel number %i.\n", j);
			cleanup( pipe_handle, test_client );
			return 1;
		}
		
		sprintf(v_jack_port_name, "Audio FIFO Port_right_ %i", j);
		audio_input_port_right[j] = jack_port_register(test_client, v_jack_port_name, 
							JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
		if (audio_input_port_right[j] == NULL) {
			fprintf(stderr, "Error: jack_port_register returned NULL for right channel number %i.\n", j);
			cleanup( pipe_handle, test_client );
			return 1;
		}
		
		
	}
	
	//Tell the jackd server what function call when it wants more audio data.
	if((error = jack_set_process_callback(test_client, process, pipe_handle)) != 0) { 
		fprintf(stderr, "Jack could not set the callback, (error %i).\n", error);
		cleanup( pipe_handle, test_client );
		return 1;
	}
	
	//Tell jack it's ok to start asking us for audio data.
	if((error = jack_activate(test_client)) != 0) {
		fprintf(stderr, "Jack could not activate the client (error %i).\n", error);
		cleanup( pipe_handle, test_client );
		return 1;
	}
	else if(verbose_flag) fprintf(stderr,"Activated client.\n");
	
	for(j=0; j<3; j++) {
		//Connect the ports.
		sprintf(v_jack_port_name, "system:capture_%i",2*j+3);/* First two ports use XLR connectors; skip them. */
		if (jack_connect (test_client, v_jack_port_name, jack_port_name (audio_input_port_left[j]))) {
	        fprintf (stderr, "cannot connect port: system:capture_%i\n", 2*j+3);/* First two ports use XLR connectors; skip them. */	
	    }
	    else if(verbose_flag) fprintf(stderr, "Connected to port: system:capture_%i\n", 2*j+3);/* First two ports use XLR connectors; skip them. */	
	    
	    sprintf(v_jack_port_name, "system:capture_%i",2*j+4);/* First two ports use XLR connectors; skip them. */	
		if (jack_connect (test_client, v_jack_port_name, jack_port_name (audio_input_port_right[j]))) {
	        fprintf (stderr, "cannot connect port: system:capture_%i\n", 2*j+4);/* First two ports use XLR connectors; skip them. */	
	    }
	    else if(verbose_flag) fprintf(stderr, "Connected to port: system:capture_%i\n", 2*j+4);/* First two ports use XLR connectors; skip them. */	
	}
	while(getchar() != EOF){
		//Quit on ctrl-D.
	}
	
	cleanup( pipe_handle, test_client );
	
	return 0;
}


void
jack_shutdown (void *arg)
{
	fprintf (stderr, "JACK shutdown.\n");
	abort();
}

